package org.processmining.streamilpminer.algorithms;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.naming.OperationNotSupportedException;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetFactory;
import org.processmining.eventstream.core.interfaces.XSEvent;
import org.processmining.eventstream.readers.abstr.AbstractXSEventReader;
import org.processmining.eventstream.readers.acceptingpetrinet.AcceptingPetriNetXSReaderResultVisualizer;
import org.processmining.eventstream.readers.acceptingpetrinet.XSEventStreamToAcceptingPetriNetReader;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.framework.util.collection.MultiSet;
import org.processmining.hybridilpminer.enginevisitor.HVMFromToConstraintsVisitorImpl;
import org.processmining.hybridilpminer.enginevisitor.HVMPopConstraintsVisitor;
import org.processmining.hybridilpminer.enginevisitor.HVMSetupObjectiveVisitorImpl;
import org.processmining.hybridilpminer.enginevisitor.HVMSetupVariablesVisitorImpl;
import org.processmining.hybridilpminer.enginevisitor.HVMTheoryOfRegionsVisitorImpl;
import org.processmining.hybridilpminer.models.lp.variablemapping.factories.VariableMappingFactory;
import org.processmining.hybridilpminer.models.lp.variablemapping.interfaces.HybridVariableMapping;
import org.processmining.hybridilpminer.models.thr.TheoryOfRegionsConstraintImpl;
import org.processmining.hybridilpminer.utils.array.ArrayUtils;
import org.processmining.hybridilpminer.utils.lp.solution.LPSolutionUtilities;
import org.processmining.hybridilpminer.utils.trove.hashing.DoubleArrHashingStrategy;
import org.processmining.logabstractions.factories.CausalAbstractionFactory;
import org.processmining.logabstractions.models.CausalAbstraction;
import org.processmining.lpengines.factories.LPEngineFactory;
import org.processmining.lpengines.interfaces.LPEngine;
import org.processmining.lpengines.interfaces.LPEngine.EngineType;
import org.processmining.lpengines.interfaces.LPEngineVisitor;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.stream.model.datastructure.DSParameterMissingException;
import org.processmining.stream.model.datastructure.DSWrongParameterException;
import org.processmining.stream.model.datastructure.DataStructureFactory;
import org.processmining.stream.model.datastructure.PointerBasedDataStructure;
import org.processmining.stream.reader.views.XSReaderSequenceOfResultsView;
import org.processmining.streamabstractrepresentation.algorithms.UncoupledConstraintReaderImpl;
import org.processmining.streamabstractrepresentation.algorithms.UncoupledTheoryOfRegionsConstraintDFAReaderImpl;
import org.processmining.streamabstractrepresentation.parameters.UnCoupledAbstractionSchemeParametersImpl;
import org.processmining.streamilpminer.parameters.UncoupledStreamILPMinerParameters;

import com.google.common.primitives.Ints;

import gnu.trove.set.hash.TCustomHashSet;

public class UncoupledStreamILPMinerImpl
		extends AbstractXSEventReader<AcceptingPetriNet, AcceptingPetriNet, UncoupledStreamILPMinerParameters>
		implements XSEventStreamToAcceptingPetriNetReader {

	private final Set<XEventClass> activities = new HashSet<>();
	private PointerBasedDataStructure<String, TheoryOfRegionsConstraintImpl<XEventClass>> cases;

	private final UncoupledTheoryOfRegionsConstraintDFAReaderImpl<Object, UnCoupledAbstractionSchemeParametersImpl> dfaDelegate;
	private final UncoupledConstraintReaderImpl<Object, UnCoupledAbstractionSchemeParametersImpl> constraintDelegate;

	public UncoupledStreamILPMinerImpl(PluginContext context, UncoupledStreamILPMinerParameters parameters) {
		super("stream_ilp_apn_ca_" + parameters.getCaseAdministrationDataStructureType().getName() + "-"
				+ parameters.getCaseAdministrationDataStructureParameters().toString() + "aa_"
				+ parameters.getInternalRepresentationDataStructureType().getName() + "-"
				+ parameters.getInternalRepresentationDataStructureParameters().toString(),
				new XSReaderSequenceOfResultsView<>("Visualizer of: Stream ILP Miner (Accepting Petri Net)", null,
						parameters.getRefreshRate(), AcceptingPetriNetXSReaderResultVisualizer.instance(context)),
				parameters);

		((XSReaderSequenceOfResultsView<AcceptingPetriNet>) getVisualization()).setReader(this);

		try {
			cases = DataStructureFactory.createPointerDataStructure(parameters.getCaseAdministrationDataStructureType(),
					parameters.getCaseAdministrationDataStructureParameters());
		} catch (OperationNotSupportedException | DSParameterMissingException | DSWrongParameterException e) {
			e.printStackTrace();
			cases = null;
			interrupt();
		}
		dfaDelegate = new UncoupledTheoryOfRegionsConstraintDFAReaderImpl<Object, UnCoupledAbstractionSchemeParametersImpl>(
				"dfa_delegate", null, parameters, cases, false);
		constraintDelegate = new UncoupledConstraintReaderImpl<Object, UnCoupledAbstractionSchemeParametersImpl>(
				"constraint_delegate", null, parameters, cases, false);

	}

	protected AcceptingPetriNet computeCurrentResult() {
		//compute a causal graph
		CausalAbstraction<XEventClass> ca = CausalAbstractionFactory
				.constructAlphaClassicCausalAbstraction(dfaDelegate.getCurrentResult());
		// create variable mapping.
		HybridVariableMapping<XEventClass> varMap = VariableMappingFactory
				.createHybridVariableMapping(EngineType.LPSOLVE, new HashSet<XEventClass>(), new HashSet<>(activities));

		Collection<double[]> solutions = new TCustomHashSet<>(new DoubleArrHashingStrategy());

		int maxSingle = varMap.getSingleVariableIndices().length == 0 ? -1
				: Ints.max(varMap.getSingleVariableIndices());
		int maxX = varMap.getXVariableIndices().length == 0 ? -1 : Ints.max(varMap.getXVariableIndices());
		int maxY = varMap.getYVariableIndices().length == 0 ? -1 : Ints.max(varMap.getYVariableIndices());

		int cols = Math.max(varMap.getMarkingVariableLPIndex(), maxSingle);
		cols = Math.max(cols, maxX);
		cols = Math.max(cols, maxY);

		// construct a fresh engine.
		LPEngine engine = LPEngineFactory.createLPEngine(EngineType.LPSOLVE, 0, cols);
		MultiSet<TheoryOfRegionsConstraintImpl<XEventClass>> constraints = constraintDelegate.getCurrentResult();
		LPEngineVisitor variables = new HVMSetupVariablesVisitorImpl<HybridVariableMapping<XEventClass>>(varMap);
		LPEngineVisitor thrConstr = new HVMTheoryOfRegionsVisitorImpl<HybridVariableMapping<XEventClass>, XEventClass>(
				varMap, constraints);
		LPEngineVisitor obj = new HVMSetupObjectiveVisitorImpl<HybridVariableMapping<XEventClass>, XEventClass>(varMap,
				constraints);

		engine.accept(variables);
		engine.accept(thrConstr);
		engine.accept(obj);

		LPEngineVisitor pop = new HVMPopConstraintsVisitor<HybridVariableMapping<?>>(varMap, 3);
		for (Pair<XEventClass, XEventClass> causal : ca.getAllGEQThreshold()) {
			LPEngineVisitor causalConstr = new HVMFromToConstraintsVisitorImpl<HybridVariableMapping<XEventClass>, XEventClass>(
					varMap, causal.getFirst(), causal.getSecond());
			engine.accept(causalConstr);
			double[] solution = ArrayUtils.round(engine.solveNative());
			if (solution.length != 0) {
				solutions.add(solution);
			}
			engine.accept(pop);
		}
		engine.destroy();
		Pair<Petrinet, Marking> mNet = LPSolutionUtilities.constructMarkedPetrinet(varMap, solutions);
		return AcceptingPetriNetFactory.createAcceptingPetriNet(mNet.getFirst(), mNet.getSecond());
	}

	@Override
	public void handleNextPacket(XSEvent event) {
		if (canReadPacket(event)) {
			dfaDelegate.handleNextPacket(event);
			constraintDelegate.handleNextPacket(event);
			super.handleNextPacket(event);
		}

	}

	public void processNewXSEvent(String caseId, XEventClass activity) {
		activities.add(activity);
		cases.add(caseId, constraintDelegate.getNewConstraint(caseId, activity));
	}

}
